"""Common GRPC backend client."""
import logging
import time
import warnings
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, ClassVar, Dict, Generator, Iterable, Optional, Sequence, Type, Union

import grpc
from google.protobuf.empty_pb2 import Empty
from google.protobuf.message import Message
from grpc import ChannelCredentials, ssl_channel_credentials

from .decoder import GrpcDecoder
from .exceptions import ServiceNotFound
from .utils import format_message, get_call_id

_LOGGER = logging.getLogger(__name__)

Request = Optional[Union[Message, Iterable[Message]]]


class CallWrapper:
    """Wrapper for gRPC calls.

    Arguments:
        call: A gRPC method to be called.
        request: A gRPC request to be sent.
        client: A GrpcClient instance.
        timeout: Timeout for the call.
        max_retries: A maximum number of calls to be made.
        modify_result: A callable to modify the result of the gRPC method.
    """

    def __init__(self, call: Callable, request: Request, *, client: 'GrpcClient', timeout: int = None,
                 max_retries: int = 10, modify_result: Callable[[Any], Any] = None):
        self.call = call
        self.request = request
        self.client = client
        self.timeout = timeout
        self.max_retries = max_retries
        self.modify_result = modify_result
        self._call_id = get_call_id()

    def get_reply(self) -> Any:
        """Fetch and return the reply."""
        return self.call(self.request, timeout=self.timeout)

    def decode_reply(self, reply: Message) -> Any:
        """Decode reply and return decoded result or raise decoded exception."""
        # Assert the supported structure of messages
        assert ('exception' in reply.DESCRIPTOR.fields_by_name or 'data' in reply.DESCRIPTOR.fields_by_name  # nosec
                or isinstance(reply, Empty))

        modify_result = self.modify_result or (lambda x: x)

        # Decode exception first
        if 'exception' in reply.DESCRIPTOR.fields_by_name and reply.HasField('exception'):
            raise self.client.decoder.decode_exception(reply.exception)  # type: ignore[attr-defined]

        # Decode data otherwise
        if 'data' in reply.DESCRIPTOR.fields_by_name and reply.HasField('data'):
            return modify_result(self.client.decoder.decode(reply.data))  # type: ignore[attr-defined]

        # The reply is empty
        return modify_result(None)

    def reconnect(self) -> None:
        """Try to reconnect the client if unavailable."""
        self.client.connect()

    def result(self) -> Any:
        """Return the call result."""
        # XXX: Extract the name of the gRPC method
        method = self.call._method.decode('utf-8')  # type: ignore
        _LOGGER.debug('[%s] %s(%s)', self._call_id, method, format_message(self.request))

        retries = 0
        try:
            while True:
                time.sleep(0.1 * retries)  # Wait for increasing amount of time before each call
                try:
                    retries += 1
                    reply = self.get_reply()
                    break
                except grpc.RpcError as error:
                    if error.code() == grpc.StatusCode.UNAVAILABLE and retries < self.max_retries:
                        # Connection probably failed. Try to reconnect.
                        _LOGGER.info('[%s] Connection failed. Trying to reconnect.', self._call_id)
                        self.reconnect()
                    else:
                        raise
        except grpc.RpcError as error:
            _LOGGER.info('[%s] %s failed with %s', self._call_id, method, error)
            raise

        _LOGGER.debug('[%s] %s returned %s', self._call_id, method, format_message(reply))

        return self.decode_reply(reply)


class StreamCallWrapper(CallWrapper):
    """Wrapper for gRPC call with streaming response."""

    def decode_reply(self, reply: Iterable[Message]) -> Generator[Any, None, None]:  # type: ignore[override]
        """Decode replies one by one."""
        for item in reply:
            yield super().decode_reply(item)


class FutureCallWrapper(CallWrapper):
    """Wrapper for gRPC future call."""

    def __init__(self, call: Callable, request: Request, *, client: 'GrpcClient', timeout: int = None,
                 max_retries: int = 10, modify_result: Callable[[Any], Any] = None):
        super().__init__(call, request, client=client, timeout=timeout, max_retries=max_retries,
                         modify_result=modify_result)
        # Get a future object - start an asynchronous request
        self._future = self._call_future()

    def _call_future(self) -> grpc.Future:
        """Start asynchronous request and return a future object."""
        future_call = self.call.future  # type: ignore
        return future_call(self.request, timeout=self.timeout)

    def get_reply(self) -> Any:
        """Fetch and return the reply."""
        return self._future.result()

    def reconnect(self) -> None:
        """Try to reconnect the client if unavailable."""
        # Cancel old request
        self._future.cancel()
        # Reconnect
        super().reconnect()
        # Start a new asynchronous request.
        self._future = self._call_future()

    def __getattr__(self, name: str) -> Any:
        """Proxy all other future properties."""
        return getattr(self._future, name)


class GrpcClient(object):
    """GRPC backend client.

    Class Attributes:
        decoder_cls: A default decoder class. No arguments are provided for its instantiation.
        grpc_modules: A default list of generated _pb2_grpc modules with service stubs.
        options: Custom options for the gRPC client.

    Attributes:
        decoder: The decoder.
        grpc_modules: The list of generated _pb2_grpc modules with service stubs.
        options: The final options of the gRPC client.
            The final options are merged from class and instance options, the later take precedence.
    """

    decoder_cls: ClassVar[Type[GrpcDecoder]] = GrpcDecoder
    grpc_modules: Sequence[ModuleType] = ()
    options: Dict[str, Any] = {}

    def __init__(self, netloc: str, decoder: GrpcDecoder = None, grpc_modules: Sequence[ModuleType] = None,
                 credentials: grpc.ChannelCredentials = None, options: Dict[str, Any] = None):
        """Initialize GRPC backend client instance.

        Args:
            netloc: Network location of a gRPC server.
            decoder: Custom decoder instance.
            grpc_modules: Custom list of generated _pb2_grpc modules with service stubs.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.
            options: Custom options for the gRPC client.
        """
        self.netloc = netloc
        self.credentials = credentials
        self.decoder = decoder or self.decoder_cls()
        self.grpc_modules = grpc_modules or self.grpc_modules
        self.options = dict(self.options, **(options or {}))

        # Stubs cache
        self._stubs: Dict[str, Any] = {}

        self.connect()

    def connect(self) -> None:
        """Create channel to GPRC server."""
        if self.credentials is not None:
            self.channel = grpc.secure_channel(self.netloc, self.credentials, self.options.items())
        else:
            self.channel = grpc.insecure_channel(self.netloc, self.options.items())

    def get_stub(self, service: str) -> Any:
        """Return stub for the service.

        Stubs are cached, so for each service, there is only one stub created.
        """
        stub_name = '{}Stub'.format(service)

        if stub_name in self._stubs:
            return self._stubs[stub_name]

        for module in self.grpc_modules:
            if hasattr(module, stub_name):
                stub_cls = getattr(module, stub_name)
                stub = stub_cls(self.channel)
                self._stubs[stub_name] = stub
                return stub
        else:
            raise ServiceNotFound('Service {} could not be found'.format(stub_name))

    def call(
        self, service: str, method: str, request: Request, *,
        blocking: bool = None, max_retries: int = 10, modify_result: Callable[[Any], Any] = None
    ) -> Any:
        """Call grpc method.

        Args:
            service: Name of service, that is used to get service stub.
            method: Name of service method.
            request: Request passed as an argument to service method.
            blocking: Determines whether to use blocking or non-blocking call.
            max_retries: A maximum number of calls to be made.
            modify_result: Callable that processes decoded result and modifies it. Default is identity.
        """
        if blocking is not None:
            if blocking:
                warnings.warn("Argument blocking is deprecated and will be deleted.", DeprecationWarning)
            else:
                warnings.warn("Argument blocking is deprecated, use method call_future instread.", DeprecationWarning)
                return self.call_future(service, method, request, max_retries=max_retries, modify_result=modify_result)
        stub = self.get_stub(service)
        grpc_method = getattr(stub, method)
        wrapper = CallWrapper(grpc_method, request, client=self, max_retries=max_retries, modify_result=modify_result)
        return wrapper.result()

    def call_future(self, service: str, method: str, request: Request, *, max_retries: int = 10,
                    modify_result: Callable[[Any], Any] = None) -> FutureCallWrapper:
        """Call grpc method asynchronously.

        Args:
            service: Name of service, that is used to get service stub.
            method: Name of service method.
            request: Request passed as an argument to service method.
            max_retries: A maximum number of calls to be made.
            modify_result: Callable that processes decoded result and modifies it. Default is identity.
        """
        stub = self.get_stub(service)
        grpc_method = getattr(stub, method)
        return FutureCallWrapper(grpc_method, request, client=self, max_retries=max_retries,
                                 modify_result=modify_result)

    def call_stream(self, service: str, method: str, request: Request, *, max_retries: int = 10,
                    modify_result: Callable[[Any], Any] = None) -> Generator[Any, None, None]:
        """Call grpc method with streaming response.

        Args:
            service: Name of service, that is used to get service stub.
            method: Name of service method.
            request: Request passed as an argument to service method.
            max_retries: A maximum number of calls to be made.
            modify_result: Callable that processes decoded result and modifies it. Default is identity.
        """
        stub = self.get_stub(service)
        grpc_method = getattr(stub, method)
        wrapper = StreamCallWrapper(grpc_method, request, client=self, max_retries=max_retries,
                                    modify_result=modify_result)
        return wrapper.result()  # type: ignore[no-any-return]


def make_credentials(ssl_cert: Optional[Union[str, Path]]) -> Optional[ChannelCredentials]:
    """Create client-side credentials from SSL certificate file.

    Args:
        ssl_cert: Path to a SSL certificate file.

    Returns:
        None if ssl_cert is None.
        Credentials otherwise.

    Raises:
        OSError: If an error on SSL certificate file is encountered.
    """
    if ssl_cert is not None:
        with open(ssl_cert) as file:
            return ssl_channel_credentials(file.read())
    else:
        return None
