"""Tests for client module."""
import types
from typing import Any
from unittest import TestCase
from unittest.mock import Mock, call, patch, sentinel

import grpc
import grpc._channel
from google.protobuf.empty_pb2 import Empty
from testfixtures import LogCapture, SequenceComparison, TempDirectory

from frgal.client import CallWrapper, FutureCallWrapper, GrpcClient, StreamCallWrapper, make_credentials
from frgal.decoder import GrpcDecoder
from frgal.exceptions import ServiceNotFound, UnknownError

from .proto.example_pb2 import Reply, SimpleMessage  # type: ignore[attr-defined]

try:
    from grpc._channel import _MultiThreadedRendezvous as _Rendezvous
except ImportError:  # pragma: no cover
    from grpc._channel import _Rendezvous


class CallWrapperTest(TestCase):
    def setUp(self):
        self.client = Mock(name='client')
        self.client.decoder = GrpcDecoder()

        self.request = SimpleMessage(value='One espresso, please!')

        self.log_handler = LogCapture()
        self.addCleanup(self.log_handler.uninstall)

    def test_get_reply(self):
        a_call = Mock(name='call', return_value=sentinel.reply)
        wrapper = CallWrapper(a_call, sentinel.request, client=sentinel.client)

        self.assertEqual(wrapper.get_reply(), sentinel.reply)

        self.assertEqual(a_call.mock_calls, [call(sentinel.request, timeout=None)])

    def test_get_reply_timeout(self):
        a_call = Mock(name='call', return_value=sentinel.reply)
        wrapper = CallWrapper(a_call, sentinel.request, client=sentinel.client, timeout=sentinel.timeout)

        self.assertEqual(wrapper.get_reply(), sentinel.reply)

        self.assertEqual(a_call.mock_calls, [call(sentinel.request, timeout=sentinel.timeout)])

    def test_decode_reply(self):
        wrapper = CallWrapper(sentinel.call, sentinel.request, client=self.client)
        reply = Reply()
        reply.data.result = 'Espresso!'

        self.assertEqual(wrapper.decode_reply(reply), 'Espresso!')

    def test_decode_reply_empty(self):
        wrapper = CallWrapper(sentinel.call, sentinel.request, client=self.client)

        self.assertIsNone(wrapper.decode_reply(Empty()))

    def test_decode_reply_error(self):
        wrapper = CallWrapper(sentinel.call, sentinel.request, client=self.client)
        reply = Reply()
        reply.exception.tea_pot_error.CopyFrom(Reply.Exception.TeaPotError())

        with self.assertRaisesRegex(UnknownError, 'tea_pot_error'):
            wrapper.decode_reply(reply)

    def test_decode_reply_modify_result(self):
        modifier = Mock(name='modifier', return_value=sentinel.modified)
        wrapper = CallWrapper(sentinel.call, sentinel.request, client=self.client, modify_result=modifier)
        reply = Reply()
        reply.data.result = 'Espresso!'

        self.assertEqual(wrapper.decode_reply(reply), sentinel.modified)
        self.assertEqual(modifier.mock_calls, [call('Espresso!')])

    def test_reconnect(self):
        wrapper = CallWrapper(sentinel.call, sentinel.request, client=self.client)

        wrapper.reconnect()

        self.assertEqual(self.client.mock_calls, [call.connect()])

    def test_result(self):
        reply = Reply()
        reply.data.result = 'Espresso!'
        a_call = Mock(name='call', return_value=reply)
        a_call._method = b'make'
        wrapper = CallWrapper(a_call, self.request, client=self.client)

        self.assertEqual(wrapper.result(), 'Espresso!')

        self.assertEqual(a_call.mock_calls, [call(self.request, timeout=None)])
        self.assertEqual(len(self.log_handler.records), 2)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] make(.*)')
        self.assertRegex(self.log_handler.records[1].getMessage(), r'\[\w{4}\] make returned ')

    def test_result_grpc_error(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNKNOWN, "Unknown error")
        error = _Rendezvous(state, None, None, None)
        a_call = Mock(name='call', side_effect=error)
        a_call._method = b'make'
        wrapper = CallWrapper(a_call, self.request, client=self.client)

        with self.assertRaisesRegex(grpc.RpcError, 'status = StatusCode.UNKNOWN'):
            wrapper.result()

        self.assertEqual(a_call.mock_calls, [call(self.request, timeout=None)])

        self.assertEqual(len(self.log_handler.records), 2)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] make(.*)')
        self.assertRegex(self.log_handler.records[1].getMessage(), r'\[\w{4}\] make failed with ')

    def test_result_retry(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        a_call = Mock(name='call', side_effect=[error, reply])
        a_call._method = b'make'
        wrapper = CallWrapper(a_call, self.request, client=self.client)

        self.assertEqual(wrapper.result(), 'Espresso!')

        self.assertEqual(a_call.mock_calls, [call(self.request, timeout=None), call(self.request, timeout=None)])
        self.assertEqual(self.client.mock_calls, [call.connect()])

        self.assertEqual(len(self.log_handler.records), 3)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] make(.*)')
        self.assertRegex(self.log_handler.records[1].getMessage(), r'\[\w{4}\] Connection failed.')
        self.assertRegex(self.log_handler.records[2].getMessage(), r'\[\w{4}\] make returned ')

    def test_result_retry_failed(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        a_call = Mock(name='call', side_effect=error)
        a_call._method = b'make'
        wrapper = CallWrapper(a_call, self.request, client=self.client, max_retries=2)

        with self.assertRaisesRegex(grpc.RpcError, 'status = StatusCode.UNAVAILABLE'):
            wrapper.result()

        self.assertEqual(a_call.mock_calls, [call(self.request, timeout=None), call(self.request, timeout=None)])
        self.assertEqual(self.client.mock_calls, [call.connect()])

        self.assertEqual(len(self.log_handler.records), 3)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] make(.*)')
        self.assertRegex(self.log_handler.records[1].getMessage(), r'\[\w{4}\] Connection failed.')
        self.assertRegex(self.log_handler.records[2].getMessage(), r'\[\w{4}\] make failed with ')


class TestFutureCallWrapper(TestCase):
    def setUp(self):
        self.client = Mock(name='client')
        self.client.decoder = GrpcDecoder()

        self.request = SimpleMessage(value='One espresso, please!')

        self.log_handler = LogCapture()
        self.addCleanup(self.log_handler.uninstall)

    def test_init(self):
        a_call = Mock(name='call')
        a_call.future.return_value = sentinel.future

        FutureCallWrapper(a_call, sentinel.request, client=self.client)

        # Check the future is called immediately.
        self.assertEqual(a_call.mock_calls, [call.future(sentinel.request, timeout=None)])

    def test_init_timeout(self):
        a_call = Mock(name='call')
        a_call.future.return_value = sentinel.future

        FutureCallWrapper(a_call, sentinel.request, client=self.client, timeout=sentinel.timeout)

        # Check the future is called immediately.
        self.assertEqual(a_call.mock_calls, [call.future(sentinel.request, timeout=sentinel.timeout)])

    def test_get_reply(self):
        a_call = Mock(name='call')
        a_call.future.return_value.result.return_value = sentinel.reply
        wrapper = FutureCallWrapper(a_call, sentinel.request, client=self.client)

        self.assertEqual(wrapper.get_reply(), sentinel.reply)

        self.assertEqual(a_call.mock_calls, [call.future(sentinel.request, timeout=None), call.future().result()])
        self.assertEqual(self.client.mock_calls, [])

    def test_reconnect(self):
        a_call = Mock(name='call')
        wrapper = FutureCallWrapper(a_call, sentinel.request, client=self.client)

        wrapper.reconnect()

        # First all is on init, second on reconnect.
        calls = [call.future(sentinel.request, timeout=None), call.future().cancel(),
                 call.future(sentinel.request, timeout=None)]
        self.assertEqual(a_call.mock_calls, calls)
        self.assertEqual(self.client.mock_calls, [call.connect()])


class StreamCallWrapperTest(TestCase):
    def setUp(self):
        self.client = Mock(name='client')
        self.client.decoder = GrpcDecoder()

        self.log_handler = LogCapture()
        self.addCleanup(self.log_handler.uninstall)

    def test_decode_reply(self):
        wrapper = StreamCallWrapper(sentinel.call, sentinel.request, client=self.client)
        reply = Reply()
        reply.data.result = 'Espresso!'

        result = wrapper.decode_reply([reply])

        self.assertIsInstance(result, types.GeneratorType)
        self.assertEqual(tuple(result), ('Espresso!', ))

    def test_decode_reply_modify_result(self):
        modifier = Mock(name='modifier', return_value=sentinel.modified)
        wrapper = StreamCallWrapper(sentinel.call, sentinel.request, client=self.client, modify_result=modifier)
        reply = Reply()
        reply.data.result = 'Espresso!'

        result = wrapper.decode_reply([reply])

        self.assertIsInstance(result, types.GeneratorType)
        self.assertEqual(tuple(result), (sentinel.modified, ))
        self.assertEqual(modifier.mock_calls, [call('Espresso!')])


class TestGrpcClient(TestCase):
    """Test GrpcClient."""

    def setUp(self):
        self.grpc_module_mock = Mock(spec=('coffeeStub', ))

        self.log_handler = LogCapture()
        self.addCleanup(self.log_handler.uninstall)

    def test_connect_insecure(self):
        with patch('frgal.client.grpc.insecure_channel', return_value=sentinel.channel) as channel_mock:
            client = GrpcClient(sentinel.netloc)

        self.assertEqual(client.channel, sentinel.channel)
        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc, SequenceComparison())])

    def test_connect_insecure_options(self):
        with patch('frgal.client.grpc.insecure_channel', return_value=sentinel.channel) as channel_mock:
            client = GrpcClient(sentinel.netloc, options={'key': sentinel.value})

        self.assertEqual(client.channel, sentinel.channel)
        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc, SequenceComparison(('key', sentinel.value)))])

    def test_connect_secure(self):
        with patch('frgal.client.grpc.secure_channel', return_value=sentinel.channel) as channel_mock:
            client = GrpcClient(sentinel.netloc, credentials=sentinel.credentials)

        self.assertEqual(client.channel, sentinel.channel)
        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials, SequenceComparison())])

    def test_connect_secure_options(self):
        with patch('frgal.client.grpc.secure_channel', return_value=sentinel.channel) as channel_mock:
            client = GrpcClient(sentinel.netloc, credentials=sentinel.credentials, options={'key': sentinel.value})

        self.assertEqual(client.channel, sentinel.channel)
        self.assertEqual(channel_mock.mock_calls,
                         [call(sentinel.netloc, sentinel.credentials, SequenceComparison(('key', sentinel.value)))])

    def test_get_stub(self):
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        with patch('frgal.client.grpc.insecure_channel', return_value=sentinel.channel):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])

        self.assertEqual(client.get_stub('coffee'), sentinel.coffee_stub)
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(sentinel.channel)])

    def test_get_stub_cache(self):
        # Test stubs are cached
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        with patch('frgal.client.grpc.insecure_channel', return_value=sentinel.channel):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])

        # Get stub once
        client.get_stub('coffee')
        # Get stub twice
        self.assertEqual(client.get_stub('coffee'), sentinel.coffee_stub)
        # Check it was created only once
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(sentinel.channel)])

    def test_get_stub_unknown(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])

        with self.assertRaisesRegex(ServiceNotFound, 'Service teaStub could not be found'):
            client.get_stub('tea')

    def test_call(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = reply

        self.assertEqual(client.call('coffee', 'make', Empty()), 'Espresso!')

    def test_call_error(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
        reply = Reply()
        reply.exception.tea_pot_error.CopyFrom(Reply.Exception.TeaPotError())
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = reply

        with self.assertRaisesRegex(UnknownError, 'tea_pot_error'):
            client.call('coffee', 'make', Empty())

    def test_call_grpc_unavailable(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = [error, error, error, reply]

        with patch('frgal.client.grpc.insecure_channel') as channel_mock:
            self.assertEqual(client.call('coffee', 'make', Empty()), 'Espresso!')

        calls = [call(sentinel.netloc, SequenceComparison()), call(sentinel.netloc, SequenceComparison()),
                 call(sentinel.netloc, SequenceComparison())]
        self.assertEqual(channel_mock.mock_calls, calls)

    def test_call_grpc_unavailable_too_many_times(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = error

        with patch('frgal.client.grpc.insecure_channel') as channel_mock:
            with self.assertRaises(grpc.RpcError):
                client.call('coffee', 'make', Empty(), max_retries=2)

        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc, SequenceComparison())])

    def test_call_grpc_other_error(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNKNOWN, "Unknown error")
        error = _Rendezvous(state, None, None, None)
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = error

        with self.assertRaisesRegex(_Rendezvous, 'status = StatusCode.UNKNOWN'):
            client.call('coffee', 'make', Empty())

    def test_call_blocking_deprecated(self):
        # Test blocking argument is marked as deprecated
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = reply

        with self.assertWarnsRegex(DeprecationWarning, "Argument blocking is deprecated and will be deleted."):
            self.assertEqual(client.call('coffee', 'make', Empty(), blocking=True), 'Espresso!')

    def test_call_non_blocking(self):
        # Test deprecated usage: call(..., blocking=False)
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        future_mock = Mock(autospec=grpc.Future)
        future_mock.result.return_value = reply
        future_mock.code.return_value = 'OK'
        stub_mock.make.future.return_value = future_mock

        with self.assertWarnsRegex(DeprecationWarning,
                                   "Argument blocking is deprecated, use method call_future instread."):
            result = client.call('coffee', 'make', Empty(), blocking=False)

        self.assertIsInstance(result, FutureCallWrapper)
        self.assertEqual(result.result(), 'Espresso!')
        self.assertEqual(result.code(), 'OK')

    def test_call_modify_result(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = reply

        def modify_result(result):
            return '{} Shaken, not stirred.'.format(result)

        self.assertEqual(
            client.call('coffee', 'make', Empty(), modify_result=modify_result),
            'Espresso! Shaken, not stirred.'
        )

    def _test_call_future(self, **kwargs: Any) -> FutureCallWrapper:
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])

        result = client.call_future('coffee', 'make', sentinel.request, **kwargs)

        self.assertEqual(result.call, self.grpc_module_mock.coffeeStub.return_value.make)
        self.assertEqual(result.request, sentinel.request)

        return result

    def test_call_future(self):
        self._test_call_future()

    def test_call_future_max_retries(self):
        result = self._test_call_future(max_retries=sentinel.retries)
        self.assertEqual(result.max_retries, sentinel.retries)

    def test_call_future_modify_result(self):
        result = self._test_call_future(modify_result=sentinel.modifier)
        self.assertEqual(result.modify_result, sentinel.modifier)

    def test_call_stream(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = [reply]

        result = client.call_stream('coffee', 'make', Empty())

        self.assertIsInstance(result, types.GeneratorType)
        self.assertEqual(tuple(result), ('Espresso!', ))

    def test_decoder_default(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc)

        self.assertEqual(type(client.decoder), GrpcDecoder)

    def test_decoder_custom(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, decoder=sentinel.decoder)

        self.assertEqual(client.decoder, sentinel.decoder)

    def test_decoder_class(self):
        class TestDecoder(GrpcDecoder):
            pass

        class TestClient(GrpcClient):
            decoder_cls = TestDecoder

        with patch('frgal.client.grpc.insecure_channel'):
            client = TestClient(sentinel.netloc)

        self.assertIsInstance(client.decoder, TestDecoder)

    def test_grpc_modules_default(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc)

        self.assertEqual(client.grpc_modules, ())

    def test_grpc_modules_custom(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, grpc_modules=sentinel.grpc_modules)

        self.assertEqual(client.grpc_modules, sentinel.grpc_modules)

    def test_grpc_modules_class(self):
        class TestClient(GrpcClient):
            grpc_modules = sentinel.grpc_modules

        with patch('frgal.client.grpc.insecure_channel'):
            client = TestClient(sentinel.netloc)

        self.assertEqual(client.grpc_modules, sentinel.grpc_modules)

    def test_options_default(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc)

        self.assertEqual(client.options, {})

    def test_options_custom(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, options={'key': sentinel.value})

        self.assertEqual(client.options, {'key': sentinel.value})

    def test_options_class(self):
        class TestClient(GrpcClient):
            options = {'key': sentinel.value}

        with patch('frgal.client.grpc.insecure_channel'):
            client = TestClient(sentinel.netloc)

        self.assertEqual(client.options, {'key': sentinel.value})

    def test_options_update(self):
        class TestClient(GrpcClient):
            options = {'key': sentinel.value}

        with patch('frgal.client.grpc.insecure_channel'):
            client = TestClient(sentinel.netloc, options={'key2': sentinel.value2})

        self.assertEqual(client.options, {'key': sentinel.value, 'key2': sentinel.value2})

    def test_options_override(self):
        class TestClient(GrpcClient):
            options = {'key': sentinel.value}

        with patch('frgal.client.grpc.insecure_channel'):
            client = TestClient(sentinel.netloc, options={'key': sentinel.value2})

        self.assertEqual(client.options, {'key': sentinel.value2})


class MakeCredentialsTest(TestCase):
    def setUp(self):
        self.temp_dir = TempDirectory()
        self.temp_dir.write('ssl_cert.pem', b'SSL CERT CONTENT')

    def tearDown(self):
        self.temp_dir.cleanup()

    def test_file(self):
        with patch('frgal.client.ssl_channel_credentials') as cred_mock:
            cred_mock.return_value = sentinel.credentials

            result = make_credentials(self.temp_dir.getpath('ssl_cert.pem'))

        self.assertEqual(result, sentinel.credentials)
        self.assertEqual(cred_mock.mock_calls, [call('SSL CERT CONTENT')])

    def test_none(self):
        self.assertIsNone(make_credentials(None))
